use std::fs;
use std::path::PathBuf;
use std::process::Command;
use std::result::Result;
use structopt::StructOpt;
use ar::Archive;
use std::fs::File;
use std::io;
use std::io::Write;
use regex::Regex;

#[derive(Debug, StructOpt)]
#[structopt(name = "Localize", about = "Localize a static library into a single object archive")]
struct LocalizeArgs
{
    /// list of file to merge
    #[structopt(name = "FILE LIST", short = "i", long = "inputs")]
    input_files: Vec<PathBuf>,

    /// filename of the final library archive
    #[structopt(name = "OUT FILE", short = "o", long = "output")]
    output_file: PathBuf,
}

fn extract(input: PathBuf, output_directory: PathBuf)
    -> Result<Vec<PathBuf>, ()>
{
    println!("Extracting {:?} into {:?}", input, output_directory);

    let input_file = File::open(input)
        .map_err(|_| ())?;

    let mut file_list : Vec<PathBuf> = vec![];

    let mut archive = Archive::new(input_file);
    while let Some(entry_result) = archive.next_entry()
    {
        let mut entry = entry_result.unwrap();

        let identifier = std::str::from_utf8(
            entry.header().identifier()).unwrap();

        let mut output_filename = output_directory.clone();
        output_filename.push(identifier);

        println!(" - {:?}", output_filename);

        let mut file = File::create(&output_filename)
            .unwrap();

        file_list.push(output_filename);

        io::copy(&mut entry, &mut file)
            .map_err(|_| ())?;
    }

    Ok(file_list)
}

fn link_partial(inputs: &Vec<PathBuf>, output: &PathBuf)
    -> Result<(), ()>
{
    Command::new("ld")
        // Enable partial linking
        .arg("-r")
        .arg("-o")
        .arg(&output)
        .args(inputs)
        .output()
        .unwrap();
    Ok(())
}

fn localize_hidden(input: &PathBuf)
{
    /* On MacOSX, check
     * -[un|re]exported_symbols_list
     * -[un]exported_symbol
     * See
     * https://developer.apple.com/library/archive/technotes/tn2185/_index.html#//apple_ref/doc/uid/DTS10004200-CH1-SUBSECTION5
     * */

    Command::new("objcopy")
        .arg("--wildcard")
        .arg("--globalize-symbol=vlc_entry*")
        .arg("--localize-hidden")
        .arg("--strip-unneeded")
        .arg("--strip-symbol=vlc_module_name")
        .arg(&input)
        .output()
        .unwrap();
}

fn link(input_lib: &PathBuf, output_object: &PathBuf)
{
    let output_dir = "./output_dir/";

    let input = fs::canonicalize(input_lib)
        .expect("Input path is invalid");

    let objects = extract(input, PathBuf::from(&output_dir))
        .expect("Cannot extract library archive");

    println!("Archive objects successfully extracted");

    link_partial(&objects, &output_object).unwrap();
    localize_hidden(&output_object);
}

#[derive(Debug)]
struct PluginEntry
{
    lib: PathBuf,
    suffix: String,
}

fn generate_plugin(
    plugins: &Vec<PluginEntry>,
    output_directory: &PathBuf,
    output: &PathBuf
) {

    println!("{:?}", plugins);

    let plugins_entries : Vec<String> = plugins
        .iter()
        .map(|plugin| String::from("vlc_entry") + &plugin.suffix)
        .collect();

    let prototypes = plugins_entries
        .iter()
        .map(|entry| String::from("int ") + entry + "(vlc_set_cb, void*);\n")
        .fold(String::new(), |entries, entry| entries + &entry);

    let entries = plugins_entries
        .iter()
        .fold(String::new(), |entries, entry|
              entries + &entry + " ,\n    ");

    let sources = format!("
#include <stddef.h>
typedef int (*vlc_set_cb) (void *, void *, int, ...);
typedef int (*vlc_plugin_entry) (vlc_set_cb, void *);

#define VLC_EXPORT __attribute__((visibility(\"default\")))
#define VLC_LOCAL __attribute__((visibility(\"hidden\")))

VLC_EXPORT
const char vlc_module_name[] = \"FOO\";

{prototypes}

VLC_LOCAL
vlc_plugin_entry vlc_plugin_entries[] = {{
    {entries}
    NULL,
}};

VLC_EXPORT
int vlc_entry(vlc_set_cb func_set, void *opaque)
{{
    for(vlc_plugin_entry *entry=vlc_plugin_entries;
        *entry != NULL; entry++)
    {{
        int ret = (*entry)(func_set, opaque);
        if (ret != 0)
            return ret;
    }}
    return 0;
}}

VLC_EXPORT
const char * vlc_entry_license()
{{
    return {license};
}}

VLC_EXPORT
const char * vlc_entry_copyright()
{{
    return {copyright};
}}

VLC_EXPORT
const char * vlc_entry_api_version()
{{
    return {api_version};
}}
    ",
    prototypes = prototypes,
    entries = entries,
    license = "\"none\"",
    copyright = "\"none\"",
    api_version = "\"4.0.0\"");

    println!("Generated:\n{}", sources);

    let mut main_c_filename = output_directory.clone();
    main_c_filename.push("main.c");

    let mut main_o_filename = output_directory.clone();
    main_o_filename.push("main.o");

    let mut file = File::create(&main_c_filename)
        .expect("cannot create temporary file");

    file.write(&sources.as_bytes())
        .expect("cannot write source to temporary file");

    let mut build = Command::new("gcc");
    build
        .arg("-fPIC")
        .arg("-c")
        .arg("-o")
        .arg(&main_o_filename)
        .arg(&main_c_filename)
        .status().unwrap();

    let mut build = Command::new("gcc");
    build
        .arg("-shared")
        .arg("-Wl,-soname,liblugin_merged.so.1")
        .arg("-fPIC")
        .arg("-o")
        .arg(output)
        .arg(main_o_filename)
        .args(plugins.iter().map(|object| &object.lib)
              .collect::<Vec<_>>())
        .arg("/home/alexandre/workspace/vlc/worktree/static-build/build-native/src/.libs/libvlccore.a")
        .arg("/home/alexandre/workspace/vlc/worktree/static-build/build-native/compat/.libs/libcompat.a")
        .arg("-pthread")
        .arg("-lidn")
        .arg("-lm");

    println!("{:?}", build);

    build
        .status()
        .unwrap();
}

fn find_vlc_suffix(input_lib: &PathBuf)
    -> Option<String>
{
    let ret = Command::new("objdump")
        .arg("-t")
        .arg(&input_lib)
        .output()
        .unwrap();

    let vlc_entry_pattern = Regex::new(r"vlc_entry__([0-9a-zA-Z_]+)")
        .unwrap();

    let entry = String::from_utf8(ret.stdout)
        .unwrap()
        .lines()
        .inspect(|line| println!("Analysing {}", line))
        .filter_map(|line| vlc_entry_pattern.captures(line))
        .inspect(|line| println!("Found {:?}", line))
        .map(|cap| String::from(&cap[0]["vlc_entry__".len()..]))
        .next();

    return entry
}


fn main() {
    let opt = LocalizeArgs::from_args();
    println!("{:?}", opt);

    let output_directory = "./output_dir/";

    // TODO: the output directory might already exist
    fs::create_dir(&output_directory);
        //.expect("Cannot create output directory");


    // Intermediate files will be generated in the following directory.
    let output_dir = fs::canonicalize(output_directory)
        .unwrap();

    // Prepare one output object path for each input plugin library.
    let output_files = opt.input_files
        .iter()
        .map(|input_file| PathBuf::from(&output_dir)
             .join(format!("{}.o", input_file.file_name().unwrap().to_string_lossy())))
        .collect::<Vec<_>>();

    // Pre-link each library to put each of them in a single object
    // file and localize the private symbols.
    opt.input_files
        .iter()
        .zip(output_files.iter())
        .for_each(|(input_lib, output_object)| link(&input_lib, &output_object));

    // Generate the plugin entries information needed for the new
    // merging plugin.
    let plugin_entries = opt.input_files
        .iter()
        .map(|filepath| find_vlc_suffix(&filepath).unwrap())
        .zip(output_files.into_iter())
        .map(|(input, object)| PluginEntry {
            suffix: format!("__{}", input),
            lib: object, })
        .collect();

    // Generate and compile the final plugin into a shared object.
    generate_plugin(
        &plugin_entries,
        &output_dir,
        &PathBuf::from(&opt.output_file));
}
